import gampixpy
from gampixpy import units, mobility

import os
import yaml
import numpy as np

class Config (dict):
    def __init__(self, config_filename):
        self.config_filename = config_filename

        self.parse_config()
        self.compute_derived_parameters()
        
    def parse_config(self):
        # parse a config (yaml) file and store values
        # internally as a dict
        with open(self.config_filename) as config_file:
            self.update(yaml.load(config_file, Loader = yaml.FullLoader).items())

        # where quantities with units are specified,
        # resolve them to the internal unit system
        self.update(self.resolve_units(self))

    def resolve_units(self, sub_dict):
        if 'value' in sub_dict and 'unit' in sub_dict:
            numerical_value = sub_dict['value']
            unit = units.unit_parser(sub_dict['unit'])
            resolved_dict = numerical_value*unit
        else:
            resolved_dict = {}
            for key, value in sub_dict.items():
                if type(value) == dict:
                    resolved_dict[key] = self.resolve_units(value)
                else:
                    resolved_dict[key] = value

        return resolved_dict

class DetectorConfig (Config):
    def compute_derived_parameters(self):
        # compute any required parameters from
        # those specified in the YAML

        return

class PhysicsConfig (Config):
    def compute_derived_parameters(self):
        mobility_model = mobility.MobilityModel(self)
        self['charge_drift'].update(mobility_model.compute_parameters())

        return

class ReadoutConfig (Config):
    def compute_derived_parameters(self):
        # compute any required parameters from
        # those specified in the YAML

        # need to handle the case where the pitch doesn't evenly divide the span of the anode
        self['n_pixels_x'] = int((self['anode']['x_upper_bound'] - self['anode']['x_lower_bound'])/self['pixels']['pitch'])
        self['n_pixels_y'] = int((self['anode']['y_upper_bound'] - self['anode']['y_lower_bound'])/self['pixels']['pitch'])

        self['n_tiles_x'] = int((self['anode']['x_upper_bound'] - self['anode']['x_lower_bound'])/self['coarse_tiles']['pitch'])
        self['n_tiles_y'] = int((self['anode']['y_upper_bound'] - self['anode']['y_lower_bound'])/self['coarse_tiles']['pitch'])

        self['tile_volume_edges'] = (np.linspace(self['anode']['x_lower_bound'],
                                                 self['anode']['x_upper_bound'],
                                                 self['n_tiles_x']+1),
                                     np.linspace(self['anode']['y_lower_bound'],
                                                 self['anode']['y_upper_bound'],
                                                 self['n_tiles_y']+1))
        return

default_detector_params = DetectorConfig(os.path.join(gampixpy.__path__[0],
                                                      'detector_config',
                                                      'default.yaml'))
default_physics_params = PhysicsConfig(os.path.join(gampixpy.__path__[0],
                                                    'physics_config',
                                                    'default.yaml'))
default_readout_params = ReadoutConfig(os.path.join(gampixpy.__path__[0],
                                                    'readout_config',
                                                    'default.yaml'))



import h5py
import numpy as np

from gampixpy.readout_objects import coarse_tile_dtype, pixel_dtype
from gampixpy.input_parsing import meta_dtype

class OutputManager:
    def __init__(self, output_filename):
        self.output_filename = output_filename

        self.outfile = h5py.File(output_filename, 'w')

        self.outfile.create_dataset('coarse_hits',
                                    shape = (0,),
                                    dtype = coarse_tile_dtype,
                                    maxshape = (None,))
        self.outfile.create_dataset('pixel_hits',
                                    shape = (0,),
                                    dtype = pixel_dtype,
                                    maxshape = (None,))
        self.outfile.create_dataset('meta',
                                    shape = (0,),
                                    dtype = meta_dtype,
                                    maxshape = (None,))
        self.n_tracks = 0 # track how many tracks have been written so far

    def add_entry(self, track, meta, event_id = None):
        self.add_track(track, event_id)
        self.add_meta(meta, event_id)

        self.n_tracks += 1
        
    def add_track(self, track, event_id = None):
        coarse_tile_sample_array, pixel_sample_array = track.to_array()

        if event_id:
            coarse_tile_sample_array[:]['event id'] = event_id
            pixel_sample_array[:]['event id'] = event_id
        else:
            coarse_tile_sample_array[:]['event id'] = self.n_tracks
            pixel_sample_array[:]['event id'] = self.n_tracks
        
        n_coarse_hits = coarse_tile_sample_array.shape[0]
        n_coarse_hits_prev = self.outfile['coarse_hits'].shape[0]

        n_pixel_hits = pixel_sample_array.shape[0]
        n_pixel_hits_prev = self.outfile['pixel_hits'].shape[0]

        self.outfile['coarse_hits'].resize((n_coarse_hits+n_coarse_hits_prev,))
        self.outfile['coarse_hits'][n_coarse_hits_prev:] = coarse_tile_sample_array

        self.outfile['pixel_hits'].resize((n_pixel_hits+n_pixel_hits_prev,))
        self.outfile['pixel_hits'][n_pixel_hits_prev:] = pixel_sample_array

    def add_meta(self, meta, event_id):
        if event_id:
            meta[:]['event id'] = event_id
        else:
            meta[:]['event id'] = self.n_tracks
        
        self.outfile['meta'].resize((self.n_tracks + 1,))
        self.outfile['meta'][self.n_tracks:] = meta



import numpy as np
import torch

from gampixpy import tracks
from gampixpy.input_parsing import meta_dtype

class Generator:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        while True:
            yield self.get_sample(), self.get_meta()
    
class PointSource (Generator):
    def generate_sample_params(self):
        self.n_samples_per_point = 100000

        self.x_init = self.kwargs['x_range'][0] + (self.kwargs['x_range'][1] - self.kwargs['x_range'][0])*np.random.random()
        self.y_init = self.kwargs['y_range'][0] + (self.kwargs['y_range'][1] - self.kwargs['y_range'][0])*np.random.random()
        self.z_init = self.kwargs['z_range'][0] + (self.kwargs['z_range'][1] - self.kwargs['z_range'][0])*np.random.random()
        self.t_init = self.kwargs['t_range'][0] + (self.kwargs['t_range'][1] - self.kwargs['t_range'][0])*np.random.random()
        self.q_init = self.kwargs['q_range'][0] + (self.kwargs['q_range'][1] - self.kwargs['q_range'][0])*np.random.random()

    def get_sample(self):
        self.generate_sample_params()

        charge_4vec = torch.tensor(self.n_samples_per_point*[[self.x_init,
                                                              self.y_init,
                                                              self.z_init,
                                                              self.t_init,
                                                              ]])
        charge_values = torch.tensor(self.n_samples_per_point*[self.q_init/self.n_samples_per_point,
                                                               ])                                 
        
        return tracks.Track(charge_4vec, charge_values)

    def get_meta(self):
        meta_array = np.array([(0, 0,
                                self.q_init,
                                self.x_init,
                                self.y_init,
                                self.z_init,
                                0,
                                0,
                                0)],
                              dtype = meta_dtype)
                
        return meta_array 

class LineSource (Generator):
    def generate_sample_params(self):
        self.n_samples = 10000

        self.x_init = self.kwargs['x_range'][0] + (self.kwargs['x_range'][1] - self.kwargs['x_range'][0])*np.random.random()
        self.y_init = self.kwargs['y_range'][0] + (self.kwargs['y_range'][1] - self.kwargs['y_range'][0])*np.random.random()
        self.z_init = self.kwargs['z_range'][0] + (self.kwargs['z_range'][1] - self.kwargs['z_range'][0])*np.random.random()
        self.t_init = self.kwargs['t_range'][0] + (self.kwargs['t_range'][1] - self.kwargs['t_range'][0])*np.random.random()
        self.q = self.kwargs['q_range'][0] + (self.kwargs['q_range'][1] - self.kwargs['q_range'][0])*np.random.random()

        # maybe direction parameters can be passed as kwargs?
        # for now, let's just always use a uniform distribution
        # over the unit sphere
        self.theta = 2*torch.pi*torch.rand(1)
        self.phi = torch.arccos(1 - 2*torch.rand(1))

        self.length = self.kwargs['length_range'][0] + (self.kwargs['length_range'][1] - self.kwargs['length_range'][0])*torch.rand(1)

    def do_point_sampling(self, start_4vec, end_4vec,
                          dx, charge_per_sample):
        # point sampling with a fixed number of samples per length
        # it may be faster to do sampling another way (test in future!)
        #  - sample with fixed amount of charge
        #  - sample with fixed number of samples per segment

        segment_interval = end_4vec - start_4vec

        sample_parametric_distance = torch.linspace(0, 1, self.n_samples)
        sample_4vec = start_4vec + segment_interval*sample_parametric_distance[:,None]
        
        sample_charges = charge_per_sample*torch.ones(self.n_samples)

        return sample_4vec, sample_charges
        
    def get_sample(self):
        self.generate_sample_params()

        start_4vec = torch.tensor((self.x_init,
                                   self.y_init,
                                   self.z_init,
                                   self.t_init,
                                   ))
        dir_4vec = torch.tensor([torch.cos(torch.tensor(self.theta))*torch.sin(torch.tensor(self.phi)),
                                 torch.sin(torch.tensor(self.theta))*torch.sin(torch.tensor(self.phi)),
                                 torch.cos(torch.tensor(self.phi)),
                                 0,
                                 ])
        end_4vec = start_4vec + dir_4vec*self.length

        displacement = start_4vec[:3] - end_4vec[:3]
        dx = torch.sum(displacement**2)
        dQ = self.q/self.n_samples
        charge_4vec, charge_values = self.do_point_sampling(start_4vec,
                                                            end_4vec,
                                                            dx, dQ,
                                                            )

        return tracks.Track(charge_4vec, charge_values)

    def get_meta(self):
        meta_array = np.array([(0, 0,
                                self.q,
                                self.x_init,
                                self.y_init,
                                self.z_init,
                                self.theta,
                                self.phi,
                                self.length)],
                              dtype = meta_dtype)
                
        return meta_array 



import numpy as np

coarse_tile_dtype = np.dtype([("event id", "u4"),
                              ("tile x", "f4"),
                              ("tile y", "f4"),
                              ("hit z", "f4"),
                              ("hit charge", "f4"),
                              ],
                             align = True)
pixel_dtype = np.dtype([("event id", "u4"),
                        ("pixel x", "f4"),
                        ("pixel y", "f4"),
                        ("hit z", "f4"),
                        ("hit charge", "f4"),
                        ],
                       align = True)

class PixelSample:
    def __init__(self,
                 pixel_pos,
                 # pixel_ind,
                 hit_timestamp,
                 hit_depth,
                 hit_measurement):
        self.pixel_pos = pixel_pos
        # self.pixel_ind = pixel_ind
        self.hit_timestamp = hit_timestamp
        self.hit_depth = hit_depth
        self.hit_measurement = hit_measurement

class CoarseGridSample:
    def __init__(self,
                 coarse_cell_pos,
                 # coarse_cell_ind,
                 measurement_time,
                 measurement_depth,
                 coarse_cell_measurement):
        self.coarse_cell_pos = coarse_cell_pos
        # self.coarse_cell_ind = coarse_cell_ind
        self.coarse_measurement_time = measurement_time
        self.coarse_measurement_depth = measurement_depth
        self.coarse_cell_measurement = coarse_cell_measurement



import math
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

import SLACplots

class EventDisplay:
    MAX_POINTS_PLOTTED = 1e3
    def __init__(self, track):
        self.track_object = track
        self.init_fig()

        # self.coarse_tile_hit_kwargs = dict(facecolors = 'cyan',
        self.coarse_tile_hit_kwargs = dict(facecolors=SLACplots.stanford.palo_verde,
                                           linewidths=1,
                                           edgecolors=SLACplots.SLACblue,
                                           alpha = 0.25)
        self.pixel_hit_kwargs = dict(facecolors=SLACplots.stanford.illuminating,
                                     linewidths=1,
                                     edgecolors=SLACplots.stanford.illuminating,
                                     alpha = 0.5)
        
    def init_fig(self):
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection = '3d')

    def remove_guidelines(self):
        self.ax.axis('off')
        
    def equal_aspect(self):
        extents = np.array([getattr(self.ax, 'get_{}lim'.format(dim))() for dim in 'xy'])
        sz = extents[:,1] - extents[:,0]
        centers = np.mean(extents, axis=1)
        maxsize = max(abs(sz))
        r = maxsize/2
        for ctr, dim in zip(centers, 'xy'):
            getattr(self.ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)
            
    def show(self):
        self.equal_aspect()
        plt.show()
        
    def save(self, outfile):
        self.equal_aspect()
        self.fig.savefig(outfile)

    def draw_box(self, cell_center_xy, cell_center_z, cell_pitch, cell_height, **kwargs):
        x_bounds = [cell_center_xy[0] - 0.5*cell_pitch,
                    cell_center_xy[0] + 0.5*cell_pitch]
        y_bounds = [cell_center_xy[1] - 0.5*cell_pitch,
                    cell_center_xy[1] + 0.5*cell_pitch]
        z_bounds = [cell_center_z,
                    cell_center_z + cell_height]
            
        bottom_face = np.array([[x_bounds[0], x_bounds[0],
                                 x_bounds[1], x_bounds[1], x_bounds[0]],
                                [y_bounds[0], y_bounds[1],
                                 y_bounds[1], y_bounds[0], y_bounds[0]],
                                [z_bounds[0], z_bounds[0],
                                 z_bounds[0], z_bounds[0], z_bounds[0]]]).T
        top_face = np.array([[x_bounds[0], x_bounds[0],
                              x_bounds[1], x_bounds[1], x_bounds[0]],
                             [y_bounds[0], y_bounds[1],
                              y_bounds[1], y_bounds[0], y_bounds[0]],
                             [z_bounds[1], z_bounds[1],
                              z_bounds[1], z_bounds[1], z_bounds[1]]]).T
        left_face = np.array([[x_bounds[0], x_bounds[0],
                               x_bounds[0], x_bounds[0], x_bounds[0]],
                              [y_bounds[0], y_bounds[1],
                               y_bounds[1], y_bounds[0], y_bounds[0]],
                              [z_bounds[0], z_bounds[0],
                               z_bounds[1], z_bounds[1], z_bounds[0]]]).T
        right_face = np.array([[x_bounds[1], x_bounds[1],
                                x_bounds[1], x_bounds[1], x_bounds[1]],
                               [y_bounds[0], y_bounds[1],
                                y_bounds[1], y_bounds[0], y_bounds[0]],
                               [z_bounds[0], z_bounds[0],
                                z_bounds[1], z_bounds[1], z_bounds[0]]]).T
        back_face = np.array([[x_bounds[0], x_bounds[1],
                               x_bounds[1], x_bounds[0], x_bounds[0]],
                              [y_bounds[0], y_bounds[0],
                               y_bounds[0], y_bounds[0], y_bounds[0]],
                              [z_bounds[0], z_bounds[0],
                               z_bounds[1], z_bounds[1], z_bounds[0]]]).T
        front_face = np.array([[x_bounds[0], x_bounds[1],
                                x_bounds[1], x_bounds[0], x_bounds[0]],
                               [y_bounds[1], y_bounds[1],
                                y_bounds[1], y_bounds[1], y_bounds[1]],
                               [z_bounds[0], z_bounds[0],
                                z_bounds[1], z_bounds[1], z_bounds[0]]]).T
        faces = [bottom_face,
                 top_face,
                 left_face,
                 right_face,
                 back_face,
                 front_face,
                 ]
        
        self.ax.add_collection3d(Poly3DCollection(faces, **kwargs))            

    def plot_raw_track(self, **kwargs):

        n_points = self.track_object.raw_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            xs = self.track_object.raw_track['position'][::reduction_factor,0]
            ys = self.track_object.raw_track['position'][::reduction_factor,1]
            zs = self.track_object.raw_track['position'][::reduction_factor,2]
            colors = np.log(self.track_object.raw_track['charge'][::reduction_factor])
        else:
            xs = self.track_object.raw_track['position'][:,0],
            ys = self.track_object.raw_track['position'][:,1],
            zs = self.track_object.raw_track['position'][:,2],
            colors = np.log(self.track_object.raw_track['charge'][:])
            
        self.ax.scatter(xs, ys, zs,
                        c = colors,
                        **kwargs,
                        )

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')
            
    def plot_drifted_track(self, **kwargs):

        n_points = self.track_object.drifted_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            xs = self.track_object.drifted_track['position'][::reduction_factor,0]
            ys = self.track_object.drifted_track['position'][::reduction_factor,1]
            zs = self.track_object.drifted_track['position'][::reduction_factor,2]
            colors = np.log(self.track_object.drifted_track['charge'][::reduction_factor])
        else:
            xs = self.track_object.drifted_track['position'][:,0]
            ys = self.track_object.drifted_track['position'][:,1]
            zs = self.track_object.drifted_track['position'][:,2]
            colors = np.log(self.track_object.drifted_track['charge'][:])
            
        self.ax.scatter(xs, ys, zs,
                        c = colors,
                        **kwargs,
                        )

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_drifted_track_timeline(self, **kwargs):

        n_points = self.track_object.drifted_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            xs = self.track_object.drifted_track['position'][::reduction_factor,0]
            ys = self.track_object.drifted_track['position'][::reduction_factor,1]
            zs = self.track_object.drifted_track['times'][::reduction_factor]
            colors = np.log(self.track_object.drifted_track['charge'][::reduction_factor])
        else:
            xs = self.track_object.drifted_track['position'][:,0]
            ys = self.track_object.drifted_track['position'][:,1]
            zs = self.track_object.drifted_track['times'][:]
            colors = np.log(self.track_object.drifted_track['charge'][:])
                        

        self.ax.scatter(xs, ys, zs,
                        c = colors,
                        **kwargs,
                        )

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')
        
    def plot_coarse_tile_measurement(self, readout_config):
        for this_hit in self.track_object.coarse_tiles_samples:
            cell_center_xy = this_hit.coarse_cell_pos
            # cell_center_z = this_hit.coarse_measurement_time
            cell_center_t = this_hit.coarse_measurement_time
            v = 1.6e5
            cell_center_z = cell_center_t*v
            cell_hit_length = v*readout_config['coarse_tiles']['clock_interval']*readout_config['coarse_tiles']['integration_length']
            cell_measurement = this_hit.coarse_cell_measurement

            self.draw_box(cell_center_xy, cell_center_z,
                          readout_config['coarse_tiles']['pitch'],
                          cell_hit_length,
                          **self.coarse_tile_hit_kwargs)
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_coarse_tile_measurement_timeline(self, readout_config):
        xlim = []
        ylim = []
        zlim = []
        
        for this_hit in self.track_object.coarse_tiles_samples:
            cell_center_xy = this_hit.coarse_cell_pos
            # cell_center_z = this_hit.coarse_measurement_time
            cell_trigger_t = this_hit.coarse_measurement_time
            cell_measurement = this_hit.coarse_cell_measurement
            cell_hit_length = readout_config['coarse_tiles']['clock_interval']*readout_config['coarse_tiles']['integration_length']

            self.draw_box(cell_center_xy, cell_trigger_t,
                          readout_config['coarse_tiles']['pitch'],
                          cell_hit_length,
                          **self.coarse_tile_hit_kwargs)
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'Arrival Time [us]')
        
    def plot_pixel_measurement(self, readout_config):
        
        for this_hit in self.track_object.pixel_samples:
            cell_center_xy = this_hit.pixel_pos
            cell_center_t = this_hit.hit_timestamp
            v = 1.6e5
            cell_center_z = cell_center_t*v
            cell_measurement = this_hit.hit_measurement
            cell_hit_length = v*readout_config['pixels']['clock_interval']*readout_config['pixels']['integration_length']
            
            self.draw_box(cell_center_xy, cell_center_z,
                          readout_config['pixels']['pitch'],
                          cell_hit_length,
                          **self.pixel_hit_kwargs)
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')

    def plot_pixel_measurement_timeline(self, readout_config):
        
        for this_hit in self.track_object.pixel_samples:
            cell_center_xy = this_hit.pixel_pos
            # cell_center_z = this_hit.hit_timestamp
            cell_trigger_t = this_hit.hit_timestamp
            cell_measurement = this_hit.hit_measurement
            cell_hit_length = readout_config['pixels']['clock_interval']*readout_config['pixels']['integration_length']
            
            self.draw_box(cell_center_xy, cell_trigger_t,
                          readout_config['pixels']['pitch'],
                          cell_hit_length,
                          **self.pixel_hit_kwargs)
            
        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'arrival time [us]')



import torch

class RecombinationModel:
    def __init__(self, physics_config):
        self.physics_config = physics_config
        self.init_params()

    def __call__(self, *args, **kwargs):
        return self.recombination_factor(*args, **kwargs)

class BoxRecombinationModel (RecombinationModel):
    def init_params(self):
        self.E_field = self.physics_config['charge_drift']['drift_field']
        self.LAr_density = self.physics_config['material']['density']

        self.alpha = self.physics_config['box_model']['box_alpha']
        self.beta = self.physics_config['box_model']['box_beta']
        
    def recombination_factor(self, dE, dx, dEdx, **kwargs):
        xi = self.beta*dEdx/(self.E_field*self.LAr_density)
        recomb = torch.max(torch.stack([torch.zeros_like(dE),
                                        torch.log(self.alpha + xi)/xi]),
                           dim = 0)[0]

        return recomb

class BirksRecombinationModel (RecombinationModel):
    def init_params(self):
        self.E_field = self.physics_config['charge_drift']['drift_field']
        self.LAr_density = self.physics_config['material']['density']

        self.ab = self.physics_config['birks_model']['birks_ab']
        self.kb = self.physics_config['birks_model']['birks_kb']
        
    def recombination_factor(self, dE, dx, dEdx, **kwargs):
        recomb = torch.max(torch.stack([torch.zeros_like(dE),
                                        self.ab/(1 + self.kb*dEdx/(self.E_field*self.LAr_density))]),
                           dim = 0)[0]

        return recomb



import numpy as np

from gampixpy.readout_objects import coarse_tile_dtype, pixel_dtype

class Track:
    def __init__(self, sample_4vec, sample_charges):
        self.raw_track = {'4vec': sample_4vec,
                          'charge': sample_charges}

        self.drifted_track = {}

        self.pixel_samples = {}
        self.coarse_tiles_samples = {}

    def to_array(self):
        coarse_tile_sample_array = np.array([(0,
                                              hit.coarse_cell_pos[0],
                                              hit.coarse_cell_pos[1],
                                              hit.coarse_measurement_time,
                                              hit.coarse_cell_measurement)
                                             for hit in self.coarse_tiles_samples],
                                            dtype = coarse_tile_dtype)
        pixel_sample_array = np.array([(0,
                                        hit.pixel_pos[0],
                                        hit.pixel_pos[1],
                                        hit.hit_timestamp,
                                        hit.hit_measurement)
                                       for hit in self.pixel_samples],
                                      dtype = pixel_dtype)

        return coarse_tile_sample_array, pixel_sample_array

# this seems to be relevant only for specific inputs
# let's not use this for now
# class Segment:
#     def __init__(self, start_position, end_position, energy):
#         self.start_position = start_position
#         self.end_position = end_position
#         self.ionization_energy = energy
